"""
An implementation of Pedersen commitment and utility support for using it.
Features: Commitment, opening, nzkp of opening

Example:
    >>> values = [Bn(2651), Bn(1), Bn(98)]
    >>> pparam = CommitParam(hs_size=len(values))
    >>> pcommit, prand = pparam.commit(values)
    >>> valid = pcommit.verify(pparam, prand, values)
    >>> print(valid)
    True
    >>> proof = pcommit.prove_knowledge(pparam, prand, values)
    >>> valid = pcommit.verify_proof(pparam, proof)
    >>> print(valid)
    True

"""

from __future__ import annotations

from hashlib import sha256
from typing import (
    Collection,
    List,
    Optional,
    Tuple,
)

import attr

from petlib.ec import EcGroup, EcPt
from petlib.bn import Bn

from . import config


class PedersenParameters:
    """ Common parameters for Pedersen commitment."""

    def __init__(self, group: Optional[EcGroup] = None, hs_size: int = 0):
        """ Return the common parameters for the specified curve.

        Args:
            group: the elliptic curve used in the scheme.
            hs_size: generate $hs_size bases for values. expand_params_len
                allows expanding parameters.
        """
        self.group: EcGroup = EcGroup(config.DEFAULT_GROUP_ID) if group is None else group
        self.q: Bn = self.group.order()

        self.H: EcPt = self.group.hash_to_point(b"com_h")
        self.HS: List[EcPt] = list()

        self.expand_params_len(hs_size)


    def num_params(self) -> int:
        return len(self.HS)


    def __repr__(self) -> str:
        return f"<{type(self).__name__}:{self.__dict__}"


    def expand_params_len(self, hs_size: int) -> None:
        """Expand the number of available bases."""
        if hs_size <= len(self.HS):
            return

        for i in range(len(self.HS), hs_size):
            self.HS.append(self.group.hash_to_point(f'com_h{i}'.encode("utf8")))


    def verify_parameters(self, valid_parameters: Optional[PedersenParameters] = None):
        """Verifies parameters. This function always checks generators'
        randomness by reproducing the hash_to_point. valid_parameters is
        optional and uses a valid parameter as the base to prevent unnecessary
        has_to_points and improve the performance.

        Args:
            valid_parameters (CommitParam): a valid Pedersen parameter
        Returns:
            boolean: pass/fail
        """

        if valid_parameters is None:
            valid_parameters = PedersenParameters()
        valid_parameters.expand_params_len(self.num_params())

        if valid_parameters.H != self.H:
            return False

        if any(valid_parameters.HS[i] != hs for i, hs in enumerate(self.HS)):
            return False

        return True


    def commit(self, values: Collection[Bn]) -> Tuple[PedersenCommitment, PedersenRandom]:
        """"commit to values.

        Args:
            values (Bn mod q): committed values.

        Returns:
            PedersenCommitment: commitment
            PedersenRand: commitment's randomness
        """

        self.expand_params_len(len(values))

        rand: Bn = self.q.random()
        C: EcPt = rand * self.H
        C += self.group.wsum(values, self.HS[:len(values)])

        return PedersenCommitment(C), PedersenRandom(rand)
    
    def extend_commit(self, old_commitment: PedersenCommitment, old_rand: PedersenRandom=None, new_values: Collection[Bn]) -> PedersenCommitment:
        """Extend a commitment with new values.

        Args:
            old_commitment (PedersenCommitment): old commitment
            new_values (Bn mod q): new values to be added to the commitment

        Returns:
            PedersenCommitment: extended commitment
        """
        self.expand_params_len(len(self.HS) + len(new_values))

        new_rands = [self.q.random() for _ in range(len(new_values))]
        C = old_commitment.commit + self.group.wsum(new_values, self.HS[len(self.HS)-len(new_values):])

        new_rand = old_rand.rand + sum(new_rands)

        return PedersenCommitment(C), PedersenRandom(new_rand)


@attr.s
class PedersenRandom:
    """Pedersen commitment's random parameter."""
    rand = attr.ib() # type: Bn


@attr.s
class PedersenProof:
    """A NIZK proof for knowing the values inside a PedersenCommitment

    Attributes:
        challenge (Bn mod q): fiat-shamir challenge,
                challenge = SHA-256(pcommit, sigma_commit)
        response (Bn mod q, Bn mod q[]): sigma protocol's response values
    """
    challenge = attr.ib() # type: Bn
    response = attr.ib()  # type: Bn


class PedersenCommitment:

    def __init__(self, commit: EcPt):
        """A Pedersen commitment.

        Attributes:
            commit (EcPt): commitment's point
        """
        self.commit: EcPt = commit


    def verify(
            self,
            pparam: PedersenParameters,
            prand: PedersenRandom,
            values: Collection[Bn]
        ) -> bool:
        """Verify the commitment.

        Args:
            pparam (CommitPram): prameters
            prand (PedersenRand): commitment's secret
            values (Bn[]): commitment's values

        Returns:
            boolean: pass/fail
        """
        if len(values) > len(pparam.HS):
            raise Exception(f"parameters does not support enough {len(values)} values")

        if not pparam.group.check_point(self.commit):
            return False
        if not (0 <= prand.rand < pparam.q):
            return False

        if any((not (0 <= v < pparam.q) for v in values)):
            return False

        rhs: EcPt = prand.rand * pparam.H
        rhs += pparam.group.wsum(values, pparam.HS[:len(values)])
        return self.commit == rhs


    def prove_knowledge(
            self,
            pparam: PedersenParameters,
            prand: PedersenRandom,
            values: Collection[Bn]
        ) -> PedersenRandom:
        """ A non-interactive proof of knowledge of opening an commitment.

        Args:
            pparam (CommitPram): prameters
            prand (PedersenRand): commitment's secret
            values (Bn[]): commitment's values

        Returns:
            (PedersenProof)
        """
        # sigma protocol's commitment phase
        r_h = pparam.q.random()
        r_vs = [pparam.q.random() for _ in range(len(values))]
        R = r_h * pparam.H
        R += pparam.group.wsum(r_vs, pparam.HS[:len(r_vs)])

        # sigma protocol's challenge: Fiat-Shamir
        chash = sha256(self.commit.export() + R.export()).digest()
        e = Bn.from_binary(chash) % pparam.q

        # sigma protocol's response phase
        s_h = r_h - prand.rand * e
        s_vs = [r - x * e for (x, r) in zip(values, r_vs)]
        return PedersenProof(e, (s_h, s_vs))


    def verify_proof(self, pparam: PedersenParameters, proof: PedersenProof) -> bool:
        """Verify a PedersenProof for this commitment.

        Args:
            proof (PedersenProof): a nizk proof for self
        Returns:
            boolean: pass/fail
        """
        sigma_commit: EcPt = proof.response[0] * pparam.H
        sigma_commit += proof.challenge * self.commit
        sigma_commit += pparam.group.wsum(
            proof.response[1],
            pparam.HS[:len(proof.response[1])]
        )
        chash: bytes = sha256(self.commit.export() + sigma_commit.export()).digest()
        h: Bn = Bn.from_binary(chash) % pparam.q

        return h == proof.challenge


def main():
    import doctest
    doctest.testmod(verbose=True)


if __name__ == '__main__':
    main()



""" Adds serialization support.

This module adds 'to_binary' and 'from_binary' to AC data classes. This module
needs to be imported to access 'to_binary' and 'from_binary' functions.
Furthermore, 'packb' and 'unpackb' enable serialization of AC and petlib
classes with msgpack protocol.

Check example.py for examples
"""

import msgpack
import petlib.pack

from zksk import Secret
from zksk.base import NIZK

from . import acl
from . import commitment
from . import blind_pedersen
from . import blind_signature

COUNTER_BASE = 20
_pack_reg = dict()


def packb(obj):
    """packs a serializable object with msgpack"""
    return msgpack.packb(obj, default=_default, use_bin_type=True)


def unpackb(data):
    """unpacks a serialized object with msgpack"""
    return msgpack.unpackb(data, ext_hook=petlib.pack.ext_hook, raw=False)


def _default(obj, enable_inherited_classes=False):
    """Overwriting petlib's default to allow inheritance between packable classes.

    This function packs obj with type(obj)'s packer. If enable_inherited_classes
    is true, the function checks for parent classed with a pack method if an
    exact match fails. enable_inherited_classes is only useful if you extend
    classes in AC without explicitly adding add_msgpack_support. Otherwise it
    reduces the performance.
    """

    # check exact type to prevent calling parent's packing when the child has
    # defined a packer.
    if type(obj) in _pack_reg:
        num, enc = _pack_reg[type(obj)]
        return msgpack.ExtType(num, enc(obj))

    # allow packing inherited classes without explicit packing method
    # this lowers the performance
    if enable_inherited_classes:
        for T in _pack_reg:
            if isinstance(obj, T):
                num, enc = _pack_reg[T]
                return msgpack.ExtType(num, enc(obj))

    return petlib.pack.default(obj)


def add_msgpack_support_slots(cls, ext, add_cls_methods=True):
    """Adds serialization support,

    Enables packing and unpacking with msgpack with 'pack.packb' and
    'pack.unpackb' methods.

    If add_method then enables equality, reading and writing for the classs.
    Specificly, adds methods:
        bytes   <- obj.to_binary()
        obj     <- cls.from_binary(bytes)
        boolean <- obj1 == obj2

    Args:
        cls: class
        ext: an unique code for the msgpack's Ext hook
    """
    def enc(obj):
        data = {key: getattr(obj, key) for key in obj.__slots__}
        kv = {}

        if hasattr(obj, 'bc_param'):
            kv['bc_param'] = obj.bc_param
        if hasattr(obj, 'pk'):
            kv['pk'] = obj.pk
        if hasattr(obj, 'z'):
            kv['z'] = obj.z
        if hasattr(obj, 'param'):
            kv['param'] = obj.param
        if hasattr(obj, 'zeta'):
            kv['zeta'] = obj.zeta
        data.update(kv)

        return packb(data)

    def dec(data):
        obj = cls.__new__(cls)
        for key, value in unpackb(data).items():
            if key != "__weakref__":
                setattr(obj, key, value)
        return obj

    def eq(a, b):
        if type(a) != type(b):
            return NotImplemented
        for ka, kb in zip(a.__slots__, b.__slots__):
            if getattr(a, ka) != getattr(b, kb):
                return False
        return True

    if add_cls_methods:
        if cls.__eq__ is object.__eq__:
            cls.__eq__ = eq
        cls.to_bytes = enc
        cls.from_bytes = staticmethod(dec)

    _pack_reg[cls] = (ext, enc)
    petlib.pack.register_coders(cls, ext, enc, dec)


def add_msgpack_support(cls, ext, add_cls_methods=True):
    """Adds serialization support,

    Enables packing and unpacking with msgpack with 'pack.packb' and
    'pack.unpackb' methods.

    If add_method then enables equality, reading and writing for the classs.
    Specificly, adds methods:
        bytes   <- obj.to_binary()
        obj     <- cls.from_binary(bytes)
        boolean <- obj1 == obj2

    Args:
        cls: class
        ext: an unique code for the msgpack's Ext hook
    """
    def enc(obj):
        return packb(obj.__dict__)

    def dec(data):
        obj = cls.__new__(cls)
        obj.__dict__.update(unpackb(data))
        return obj

    def eq(a, b):
        if type(a) != type(b):
            return NotImplemented

        return a.__dict__ == b.__dict__

    if add_cls_methods:
        if cls.__eq__ is object.__eq__:
            cls.__eq__ = eq
        cls.to_bytes = enc
        cls.from_bytes = staticmethod(dec)

    _pack_reg[cls] = (ext, enc)
    petlib.pack.register_coders(cls, ext, enc, dec)


def register_all_classes():
    # commitment
    add_msgpack_support(commitment.PedersenParameters, COUNTER_BASE+1)
    add_msgpack_support(commitment.PedersenProof, COUNTER_BASE+2)
    add_msgpack_support(commitment.PedersenCommitment, COUNTER_BASE+3)
    # blind commitment
    add_msgpack_support(blind_pedersen.BlPedersenPrivate, COUNTER_BASE+4)
    add_msgpack_support(blind_pedersen.BlPedersenProof, COUNTER_BASE+5)
    add_msgpack_support(blind_pedersen.BlindedPedersenParam, COUNTER_BASE+6)
    add_msgpack_support(blind_pedersen.BlPedersenCommitment, COUNTER_BASE+7)
    # Abe's signature
    add_msgpack_support_slots(blind_signature.AbeParam, COUNTER_BASE+8)
    add_msgpack_support_slots(blind_signature.AbePublicKey, COUNTER_BASE+9)
    add_msgpack_support_slots(blind_signature.AbePrivateKey, COUNTER_BASE+10)
    add_msgpack_support_slots(blind_signature.AbeSignature, COUNTER_BASE+11)
    add_msgpack_support_slots(
        blind_signature.SignerCommitMessage, COUNTER_BASE+12)
    add_msgpack_support_slots(
        blind_signature.SignerResponseMessage, COUNTER_BASE+13)
    # ACL
    add_msgpack_support_slots(acl.ACLParam, COUNTER_BASE+14)
    add_msgpack_support_slots(acl.ACLIssuerPrivateKey, COUNTER_BASE+15)
    add_msgpack_support_slots(acl.ACLIssuerPublicKey, COUNTER_BASE+16)
    add_msgpack_support_slots(acl.ProveAttrKnowledgeMessage, COUNTER_BASE+17)
    add_msgpack_support_slots(acl.ACLCredential, COUNTER_BASE+18)
    add_msgpack_support_slots(acl.ACLCredentialPrivate, COUNTER_BASE+19)
    # zksk
    add_msgpack_support(Secret, COUNTER_BASE+20, add_cls_methods=False)
    add_msgpack_support(NIZK, COUNTER_BASE+21, add_cls_methods=False)

    # Added later
    add_msgpack_support_slots(
        blind_signature.SignerCommitmentInternalState, COUNTER_BASE+22)
    add_msgpack_support_slots(
        blind_signature.UserBlindedChallengeInternalState, COUNTER_BASE+23)
    add_msgpack_support_slots(
        blind_signature.BlindedChallengeMessage, COUNTER_BASE+24)


register_all_classes()


def main():
    import doctest
    doctest.testmod(verbose=True)


if __name__ == '__main__':
    main()



"""
The Baldimtsi-Lysyanskaya Anonymous Credentials Light scheme

Ref: Baldimtsi, Foteini, and Anna Lysyanskaya. "Anonymous credentials light."
Proceedings of the 2013 ACM SIGSAC conference on Computer & communications
security. ACM, 2013.

A guide to ACL:
1. Prove attribute knowledge: The user commits to his attributes. She/he proves
   the correctness of this commitment to the issuer with a non-interactive zkp.

2. Preparation: The issuer generates a random generator $Z1$ based on the
   commitment and randomness $rnd$ and sends $rnd$ to the user.

3. CredentialCommitment: the issuer sends an initial commitment $C$ as part of
   an interactive sigma protocol to the user.

4. CredentialChallenge: The user blinds the commitment $C$ with the randomeness
   $t$ to get a new commitment $C'$. The user computes the fiat-shamir challenge
   $\\epsilon$ based on $C'$, blinds it to $e$, and sends the blinded challenge
   to the prover.

5. CredentialResponse: The prover sends completes the sigma protocol and sends
   the response for $(C, e)$ to the user. The user unblinds the response to form
   a signature withZ $(C', \\epsilon)$.

Steps 2 and 3 can be combined together.

Warning: Running concurrent ACL signing sessions is insecure.

How to use:
    >>> # generating keys and wrappers
    >>> issuer_priv, issuer_pk = ACLParam().generate_new_key_pair()
    >>> issuer = ACLIssuer(issuer_priv, issuer_pk)
    >>> user = ACLUser(issuer_pk)
    >>> message = "Hello world"

    >>> # Issuance
    >>> attributes = [Bn(13), "Hello", "WoRlD", "Hidden"]
    >>> attr_proof = user.prove_attr_knowledge(attributes)
    >>> com, com_intern = issuer.commit(attr_proof)
    >>> challenge, challenge_intern = user.compute_blind_challenge(com, message)
    >>> resp = issuer.respond(challenge, com_intern)
    >>> cred_private = user.compute_credential(resp, challenge_internal)

    >>> # show credential
    >>> cred = cred_private.show_credential([True, True, True, False])
    >>> assert cred.verify_credential(issuer_pk)
    >>> print(cred.get_message())
    b'Hello world'
    >>> print(cred.get_attributes())
    [13, 'Hello', 'WoRlD', None]
"""
from __future__ import annotations
from typing import Collection, Tuple, Union

import attr

from petlib.bn import Bn
from petlib.ec import EcGroup, EcPt

from .commitment import PedersenCommitment, PedersenProof
from .blind_pedersen import BlPedersenCommitment, BlPedersenPrivate, BlPedersenProof, BlindedPedersenParam
from .blind_signature import (
    AbeParam,
    AbePrivateKey,
    AbePublicKey,
    AbeSignature,
    AbeSigner,
    AbeUser,
    BlindedChallengeMessage,
    SignerCommitMessage,
    SignerCommitmentInternalState,
    SignerResponseMessage,
    UserBlindedChallengeInternalState,
)
from .config import DEFAULT_GROUP_ID


Attribute = Union[bytes, str, Bn]


class ACLAttributeProofIsInvalid(Exception):
    """The attribute proof is invalid."""


class ACLParam(AbeParam):
    """Param for ACL credential including both AbeParam and
    BlindedPedersenParam.
    """

    def __init__(self, group: EcGroup = EcGroup(DEFAULT_GROUP_ID)):
        super().__init__(group)


    def generate_new_key_pair(self) -> Tuple[ACLIssuerPrivateKey, ACLIssuerPublicKey]:
        sk = self.q.random()
        private = ACLIssuerPrivateKey(sk)
        public = private.public_key(self)
        return private, public


@attr.s(slots=True)
class ProveAttrKnowledgeMessage:
    commit = attr.ib()     # type: PedersenCommitment
    nizk_proof = attr.ib() # type: PedersenProof


class ACLIssuerPrivateKey(AbePrivateKey):
    def __init__(self, sk: Bn):
        super().__init__(sk=sk)


    def public_key(self, params: ACLParam) -> ACLIssuerPublicKey:
        """Create a public key corresponding to this private key."""
        return ACLIssuerPublicKey(params, self)


class ACLIssuerPublicKey(AbePublicKey):
    """	Use ACLParam.generate_new_key_pair to generate a fresh key pair.

    Args:
        param (ACLParam): parameters
        priv (ACLIssuerPrivateKey): issuer's private key
    """

    __slots__ = ("bc_param",)

    def __init__(self, param: ACLParam, private: ACLIssuerPrivateKey):
        super().__init__(param, private)
        # Improvement: reusing an existing param for common gens can speed up
        # the process.
        self.bc_param = BlindedPedersenParam(
            group=param.group,
            hs_size=5,
            Z=self.z,
            H_2=self.param.g
        )


    def verify_parameters(self, verify_bases: bool = False):
        """Verifies that the public key is generated correctly.

        Args:
            verify_bases (bool): if true, verifies public parameter's generators G, H
        """
        return all(
            (
                self.z == self.bc_param.Z,
                self.param.g == self.bc_param.H_2,
                super().verify_parameters(verify_bases)
            )
        )


class ACLIssuer(AbeSigner):
    """A class which handles the issuer role

    Attributes:
        private (ACLIssuerPrivateKey): issuer's private key
        public (ACLIssuerPublicKey): issuer's public key
    """

    def __init__(self, private: ACLIssuerPrivateKey, public: ACLIssuerPublicKey):
        self.private: ACLIssuerPrivateKey
        self.public: ACLIssuerPublicKey
        super().__init__(private, public)


    def commit(
            self,
            prove_attr_msg: ProveAttrKnowledgeMessage
        ) -> Tuple[SignerCommitMessage, SignerCommitmentInternalState]:
        """Checks the attribute proof and perform AbeSignature's commit phase.

        Errors:
            attribute proof is invalid
        Args:
            prove_attr_msg (ProveAttrKnowledgeMessage): the user's commitment
                to his/her values.
        Returns:
            (SignerCommitMessage)
        """

        valid_commit = prove_attr_msg.commit.verify_proof(
            pparam = self.public.bc_param,
            proof = prove_attr_msg.nizk_proof
        )
        if not valid_commit:
            raise ACLAttributeProofIsInvalid("Attribute proof is invalid.")
        
        commit = prove_attr_msg.commit.commit
        g = self.param.g
        return super()._commit(lambda rnd, commit=commit, g=g: commit + rnd * g)
    
    def extend_commit(self, old_commitment: PedersenCommitment, new_values: Collection[Attribute]) -> PedersenCommitment:
        """Extend a Pedersen commitment with new values.

        Args:
            old_commitment (PedersenCommitment): old commitment
            new_values (Collection[Bn]): new values to be added to the commitment

        Returns:
            PedersenCommitment: new, extended commitment
        """
        bc_param: BlindedPedersenParam = self.public.bc_param
        new_pcommit: PedersenCommitment = bc_param.extend_commit(old_commitment, new_values)
        return new_pcommit

    
    def abe_commit(
        self,
        commitment: PedersenCommitment
        ) -> Tuple[SignerCommitMessage, SignerCommitmentInternalState]:
        """Perform AbeSignature's commit phase.

        Args:
            commitment (PedersenCommitment): acl user's commitment to his/her values.

        Returns:
            Tuple[SignerCommitMessage, SignerCommitmentInternalState] 
        """
        commit = commitment.commit
        g = self.param.g
        return super()._commit(lambda rnd, commit=commit, g=g: commit + rnd * g)
    
    def verify_proof(
        self,
        prove_attr_msg: ProveAttrKnowledgeMessage,
        ) -> None:
        valid_commit = prove_attr_msg.commit.verify_proof(
            pparam = self.public.bc_param,
            proof = prove_attr_msg.nizk_proof
        )
        if not valid_commit:
            raise ACLAttributeProofIsInvalid("Attribute proof is invalid.")


class ACLUser(AbeUser):
    """An ACL user which receives a credential.
    This class stores a state and only supports one credential at a time.

    Attributes:
        public (ACLIssuerPublicKey): issuer's public key
    """

    __slots__ = ("user_attr_commitment", "attributes", "pcommit", "prand")

    def __init__(self, public: ACLIssuerPublicKey):
        self.public: ACLIssuerPublicKey
        super().__init__(public)


    def _compute_z1(self, rnd: Bn) -> EcPt:
        if not (isinstance(rnd, Bn) and rnd != 0 and (0 <= rnd < self.param.q)):
            raise ValueError("Invalid registration.")
        return self.user_attr_commitment.commit + rnd * self.param.g

    def compute_blind_challenge(
            self,
            commit_message: SignerCommitMessage,
            message: Union[bytes, str]
        ) -> Tuple[BlindedChallengeMessage, UserBlindedChallengeInternalState]:
        """Receive a SignerCommitMessage from the signer and start the procedure
        of getting a signature on message m from the signer.

        Args:
            commit_message (SignerCommitMessage):response from AbeSigner.commit()
            m (bytes): message to sign.
                If m is a string, then the procedure encodes it as 'utf8'
        """
        return self._compute_blind_challenge(
            commit_message,
            message,
            self._compute_z1
        )


    def prove_attr_knowledge(self, attributes: Collection[Attribute]) -> ProveAttrKnowledgeMessage:
        """ Prove the knowledge of attributes to initiate an issuance:

        Args:
            attributes (List[Union[Bn, str, bytes]]): user's attributes

        Returns:
            (ProveAttrKnowledgeMessage)
        """
        bc_param =  self.public.bc_param
        self.attributes = tuple(attributes)

        self.pcommit, self.prand = bc_param.commit(attributes)
        proof = self.pcommit.prove_knowledge(
            pparam = bc_param,
            prand = self.prand,
            values = bc_param.process_raw_values(attributes)
        )
        self.user_attr_commitment = self.pcommit
        return ProveAttrKnowledgeMessage(commit=self.pcommit, nizk_proof=proof)
    
    def extend_commit(self, old_commitment: PedersenCommitment, new_values: Collection[Attribute]) -> PedersenCommitment:
        """Extend a Pedersen commitment with new values.

        Args:
            old_commitment (PedersenCommitment): old commitment
            new_values (Collection[Bn]): new values to be added to the commitment

        Returns:
            PedersenCommitment: new, extended commitment
        """
        self.attributes += tuple(new_values)
        bc_param: BlindedPedersenParam = self.public.bc_param
        new_pcommit: PedersenCommitment = bc_param.extend_commit(old_commitment, new_values)
        return new_pcommit


    def compute_credential(
            self,
            response: SignerResponseMessage,
            challenge_state: UserBlindedChallengeInternalState
        ) -> ACLCredentialPrivate:
        """Finish the protocol and form a private credential.

        Args:
            response (SignerRespondMessage): output of ACLIssuer.respond

        Returns:
            cred_private (ACLCredentialPrivate): A private credential.
                cred_private.show_credential() creates a one time use credential
                with the private information.
        """
        sig = self.compute_signature(response, challenge_state)
        bcommit, bpriv = self.public.bc_param.blind_commit(
            raw_values=self.attributes,
            blindness_rand=challenge_state.blinder,
            h_rand=self.prand.rand
        )
        cred_private = ACLCredentialPrivate(
            signature = sig,
            bcommit = bcommit,
            bpriv = bpriv
        )
        return cred_private


class ACLCredentialPrivate:
    """An ACL credential's secret. This object can be used to generate a
    one-time use credential token

    Args:
        signature (AbeSignature): credential's signature
        bcommit (BlPedersenCommit): blind commit's public part
        bpriv (BlPedersenPrivate): blind commit's randomness
    """

    __slots__ = ("signature", "bcommit", "bpriv", "revealed")

    def __init__(self, signature: AbeSignature, bcommit: BlPedersenCommitment, bpriv: BlPedersenPrivate):
        self.signature = signature
        self.bcommit = bcommit
        self.bpriv = bpriv
        self.revealed = False


    def show_credential(self, revealed_attrs: Collection[bool]) -> ACLCredential:
        """Show the credential and reveal $revealed_attrs attributes.

        Error:
            One time use only
            revealed_attrs should have the same len as attributes.
        Args:
            revealed_attrs (boolean []): reveals attr[i] iff revealed_attrs[i]
        Returns:
            ACLCredential: a credential
        """

        if self.revealed:
            raise Exception(
                "ACL credential is one-time use only. "
                "You cannot show it twice."
            )
        self.revealed = True
        if len(revealed_attrs) != len(self.bpriv.values):
            raise Exception(
                "Revealed_attrs is a binary mask to decide which attrs "
                "should be reveal. It must have the same size as attributes."
            )

        bproof = self.bcommit.prove_values(self.bpriv, revealed_attrs)
        cred = ACLCredential(
            signature = self.signature,
            bcommit = self.bcommit,
            bc_proof = bproof
        )
        return cred


class ACLCredential:
    """An ACL credential.
    The issuer's public key is intentionally not bundled in this class to force
    the verifier to use a known correct issuer's public key.

    Attributes:
        signature (AbeSignature): credential's signature
        bcommit (BlPedersenCommit): blind commit's public part
        bc_proof (BlPedersenProof): blind commit's proof and attributes
    """

    __slots__ = ("signature", "bcommit", "bc_proof")

    def __init__(self, signature: AbeSignature, bcommit: BlPedersenCommitment, bc_proof: BlPedersenProof):
        self.signature = signature
        self.bcommit = bcommit
        self.bc_proof = bc_proof


    def verify_credential(self, issuer_pk: ACLIssuerPrivateKey) -> bool:
        """ verifies the credential. Assumes that issuer_pk and its parameters
        are verified.

        Args:
            issuer_pk (ACLIssuerPublicKey): issuer's public key.
        Returns:
            boolean: pass/fail
        """
        if not (
            issuer_pk.verify_signature(self.signature) and
            self.bcommit.verify_proof(issuer_pk.bc_param, self.bc_proof)
        ):
            return False
        return True


    def attributes(self) -> Tuple[Attribute]:
        """ get attributes.
        Returns:
            raw_attr []: raw attributes in their original format. None for
            non-revealed attributes.
        """
        return self.bc_proof.revealed_values


    def message(self) -> bytes:
        """ get message.
        Returns:
            bytes: message
        """
        return self.signature.message


def main():
    import doctest
    doctest.testmod(verbose=True)


if __name__ == "__main__":
    main()
